#include <TFT_eSPI.h>
#include <TFT_eWidget.h>
#include "Adafruit_TinyUSB.h"
#include <FS.h>

// TFT instance
TFT_eSPI tft = TFT_eSPI();
ButtonWidget btn_1 = ButtonWidget(&tft);
ButtonWidget btn_2 = ButtonWidget(&tft);

#define BUTTON_W 128
#define BUTTON_H 128

ButtonWidget* btn[] = {&btn_1, &btn_2};
uint8_t buttonCount = sizeof(btn) / sizeof(btn[0]);

#define CALIBRATION_FILE "/TouchCalData1"
#define REPEAT_CAL true

enum { RID_KEYBOARD = 1 };
uint8_t const desc_hid_report[] = { TUD_HID_REPORT_DESC_KEYBOARD(HID_REPORT_ID(RID_KEYBOARD)) };
uint8_t keycode[6] = { HID_KEY_NONE, HID_KEY_NONE, HID_KEY_NONE, HID_KEY_NONE, HID_KEY_NONE, HID_KEY_NONE };
Adafruit_USBD_HID usb_hid;

void btn_1_pressAction(void) {
  if (btn_1.justPressed()) {
    keycode[0] = HID_KEY_C;
    usb_hid.keyboardReport(RID_KEYBOARD, KEYBOARD_MODIFIER_LEFTCTRL, keycode);
    delay(10);
    usb_hid.keyboardRelease(RID_KEYBOARD);
    btn_1.drawSmoothButton(true);
  }
}

void btn_1_releaseAction(void) {
  if (btn_1.justReleased()) {
    usb_hid.keyboardRelease(RID_KEYBOARD);
    btn_1.drawSmoothButton(false);
  }
}

void btn_2_pressAction(void) {
  if (btn_2.justPressed()) {
    keycode[0] = HID_KEY_V;
    usb_hid.keyboardReport(RID_KEYBOARD, KEYBOARD_MODIFIER_LEFTCTRL, keycode);
    delay(10);
    usb_hid.keyboardRelease(RID_KEYBOARD);
    btn_2.drawSmoothButton(true);
  }
}

void btn_2_releaseAction(void) {
  if (btn_2.justReleased()) {
    usb_hid.keyboardRelease(RID_KEYBOARD);
    btn_2.drawSmoothButton(false);
  }
}

void initButtons() {
  uint16_t buttonPadding = 20;
  uint16_t startX = buttonPadding;
  uint16_t startY = buttonPadding;
  uint16_t xIncrement = BUTTON_W + buttonPadding;

  // Row 1
  btn_1.initButtonUL(startX, startY, BUTTON_W, BUTTON_H, TFT_WHITE, TFT_RED, TFT_YELLOW, "COPY", 2);
  btn_1.setPressAction(btn_1_pressAction);
  btn_1.setReleaseAction(btn_1_releaseAction);
  btn_1.drawSmoothButton(false, 3, TFT_BLACK); // 3 is outline width, TFT_BLACK is the surrounding background colour for anti-aliasing

  btn_2.initButtonUL(startX + xIncrement, startY, BUTTON_W, BUTTON_H, TFT_WHITE, TFT_RED, TFT_YELLOW, "PASTE", 2);
  btn_2.setPressAction(btn_2_pressAction);
  btn_2.setReleaseAction(btn_2_releaseAction);
  btn_2.drawSmoothButton(false, 3, TFT_BLACK); // 3 is outline width, TFT_BLACK is the surrounding background colour for anti-aliasing
}

void touch_calibrate() {
  uint16_t calData[5];
  uint8_t calDataOK = 0;

  // Check file system exists
  if (!LittleFS.begin()) {
    Serial.println("Formating file system");
    LittleFS.format();
    LittleFS.begin();
  }

  // Check if calibration file exists and size is correct
  if (LittleFS.exists(CALIBRATION_FILE)) {
    if (REPEAT_CAL) {
      // Delete if we want to re-calibrate
      LittleFS.remove(CALIBRATION_FILE);
    } else {
      File f = LittleFS.open(CALIBRATION_FILE, "r");
      if (f) {
        if (f.readBytes((char *)calData, 14) == 14) {
          calDataOK = 1;
        }
        f.close();
      }
    }
  }

  if (calDataOK && !REPEAT_CAL) {
    // Calibration data valid
    tft.setTouch(calData);
  } else {
    // Data not valid so recalibrate
    tft.fillScreen(TFT_BLACK);
    tft.setCursor(20, 0);
    tft.setTextFont(2);
    tft.setTextSize(1);
    tft.setTextColor(TFT_WHITE, TFT_BLACK);

    tft.println("Touch corners as indicated");

    tft.setTextFont(1);
    tft.println();

    if (REPEAT_CAL) {
      tft.setTextColor(TFT_RED, TFT_BLACK);
      tft.println("Set REPEAT_CAL to false to stop this running again!");
    }

    tft.calibrateTouch(calData, TFT_MAGENTA, TFT_BLACK, 15);

    tft.setTextColor(TFT_GREEN, TFT_BLACK);
    tft.println("Calibration complete!");

    // Store data
    File f = LittleFS.open(CALIBRATION_FILE, "w");
    if (f) {
      f.write((const unsigned char *)calData, 14);
      f.close();
    }
  }
}

void setup() {
  usb_hid.setPollInterval(2);
  usb_hid.setReportDescriptor(desc_hid_report, sizeof(desc_hid_report));
  usb_hid.begin();
  while (!USBDevice.mounted()) delay(1);
  do delay(50); while (!usb_hid.ready());

  tft.begin();
  tft.invertDisplay(1);
  tft.setRotation(0);
  tft.fillScreen(TFT_BLACK);

  touch_calibrate();
  tft.fillScreen(TFT_BLACK);
  initButtons();
}

void loop() {
  static uint32_t scanTime = millis();
  uint16_t t_x = 9999, t_y = 9999; // To store the touch coordinates

  // Scan keys every 50ms at most
  if (millis() - scanTime >= 50) {
    // Pressed will be set true if there is a valid touch on the screen
    bool pressed = tft.getTouch(&t_x, &t_y);
    scanTime = millis();
    for (uint8_t b = 0; b < buttonCount; b++) {
      if (pressed) {
        if (btn[b]->contains(t_x, t_y)) {
          btn[b]->press(true);
          btn[b]->pressAction();
        }
      } else {
        btn[b]->press(false);
        btn[b]->releaseAction();
      }
    }
  }
}
